import itertools

from .abstract_io import (AbstractDxlIO, _DxlControl,
                          _DxlAccess, DxlTimeoutError)
from .. import conversion as conv
from ..protocol import v1 as v1


class DxlIO(AbstractDxlIO):
    _protocol = v1

    def factory_reset(self):
        """ Reset all motors on the bus to their factory default settings. """
        try:
            self._send_packet(self._protocol.DxlResetPacket())

        except DxlTimeoutError:
            pass

    def get_control_mode(self, ids):
        """ Gets the mode ('joint' or 'wheel') for the specified motors. """
        to_get_ids = [id for id in ids if id not in self._known_mode]
        limits = self.get_angle_limit(to_get_ids, convert=False)
        modes = ('wheel' if limit == (0, 0) else 'joint' for limit in limits)

        self._known_mode.update(zip(to_get_ids, modes))

        return tuple(self._known_mode[id] for id in ids)

    def set_wheel_mode(self, ids):
        """ Sets the specified motors to wheel mode. """
        self.set_control_mode(dict(zip(ids, itertools.repeat('wheel'))))

    def set_joint_mode(self, ids):
        """ Sets the specified motors to joint mode. """
        self.set_control_mode(dict(zip(ids, itertools.repeat('joint'))))

    def set_control_mode(self, mode_for_id):
        models = ['MX' if m.startswith('MX') else '*'
                  for m in self.get_model(list(mode_for_id.keys()))]
        pos_max = [conv.position_range[m][0] for m in models]
        limits = ((0, 0) if mode == 'wheel' else (0, pos_max[i] - 1)
                  for i, mode in enumerate(mode_for_id.itervalues()))

        self._set_angle_limit(dict(zip(mode_for_id.keys(), limits)), convert=False)
        self._known_mode.update(mode_for_id.items())

    def set_angle_limit(self, limit_for_id, **kwargs):
        """ Sets the angle limit to the specified motors. """
        convert = kwargs['convert'] if 'convert' in kwargs else self._convert

        if 'wheel' in self.get_control_mode(limit_for_id.keys()):
            raise ValueError('can not change the angle limit of a motor in wheel mode')

        if (0, 0) in limit_for_id.values():
            raise ValueError('can not set limit to (0, 0)')

        self._set_angle_limit(limit_for_id, convert=convert)

# MARK: - Generate the accessors


def _add_control(name,
                 address, length=2, nb_elem=1,
                 access=_DxlAccess.readwrite,
                 models=set(conv.dynamixelModels.values()),
                 dxl_to_si=lambda val, model: val,
                 si_to_dxl=lambda val, model: val,
                 getter_name=None,
                 setter_name=None):

    control = _DxlControl(name,
                          address, length, nb_elem,
                          access,
                          models,
                          dxl_to_si, si_to_dxl,
                          getter_name, setter_name)

    DxlIO._generate_accessors(control)


_add_control('model',
             address=0x00,
             access=_DxlAccess.readonly,
             dxl_to_si=conv.dxl_to_model)

_add_control('firmware',
             address=0x02, length=1,
             access=_DxlAccess.readonly)

_add_control('id',
             address=0x03, length=1,
             access=_DxlAccess.writeonly,
             setter_name='change_id')

_add_control('baudrate',
             address=0x04, length=1,
             access=_DxlAccess.writeonly,
             setter_name='change_baudrate',
             si_to_dxl=conv.baudrate_to_dxl)

_add_control('return delay time',
             address=0x05, length=1,
             dxl_to_si=conv.dxl_to_rdt,
             si_to_dxl=conv.rdt_to_dxl)

_add_control('angle limit',
             address=0x06, nb_elem=2,
             dxl_to_si=lambda value, model: (conv.dxl_to_degree(value[0], model),
                                             conv.dxl_to_degree(value[1], model)),
             si_to_dxl=lambda value, model: (conv.degree_to_dxl(value[0], model),
                                             conv.degree_to_dxl(value[1], model)))

_add_control('drive mode',
             address=0x0A, length=1,
             access=_DxlAccess.readwrite,
             models=('MX-106', ),
             dxl_to_si=conv.dxl_to_drive_mode,
             si_to_dxl=conv.drive_mode_to_dxl)

_add_control('highest temperature limit',
             address=0x0B, length=1,
             dxl_to_si=conv.dxl_to_temperature,
             si_to_dxl=conv.temperature_to_dxl)

_add_control('voltage limit',
             address=0x0C, length=1, nb_elem=2,
             dxl_to_si=lambda value, model: (conv.dxl_to_voltage(value[0], model),
                                             conv.dxl_to_voltage(value[1], model)),
             si_to_dxl=lambda value, model: (conv.voltage_to_dxl(value[0], model),
                                             conv.voltage_to_dxl(value[1], model)))

_add_control('max torque',
             address=0x0E,
             dxl_to_si=conv.dxl_to_torque,
             si_to_dxl=conv.torque_to_dxl)

_add_control('status return level',
             address=0x10, length=1,
             dxl_to_si=conv.dxl_to_status,
             si_to_dxl=conv.status_to_dxl)

_add_control('alarm LED',
             address=0x11, length=1,
             dxl_to_si=conv.dxl_to_alarm,
             si_to_dxl=conv.alarm_to_dxl)

_add_control('alarm shutdown',
             address=0x12, length=1,
             dxl_to_si=conv.dxl_to_alarm,
             si_to_dxl=conv.alarm_to_dxl)

_add_control('torque_enable',
             address=0x18, length=1,
             dxl_to_si=conv.dxl_to_bool,
             si_to_dxl=conv.bool_to_dxl,
             getter_name='is_torque_enabled',
             setter_name='_set_torque_enable')

_add_control('LED',
             address=0x19, length=1,
             dxl_to_si=conv.dxl_to_bool,
             si_to_dxl=conv.bool_to_dxl,
             setter_name='_set_LED',
             getter_name='is_led_on')

_add_control('pid gain',
             address=0x1A, length=1, nb_elem=3,
             models=('MX-12', 'MX-28', 'MX-64', 'MX-106'),
             dxl_to_si=conv.dxl_to_pid,
             si_to_dxl=conv.pid_to_dxl)

_add_control('compliance margin',
             address=0x1A, length=1, nb_elem=2,
             models=('AX-12', 'AX-18', 'RX-24', 'RX-28', 'RX-64'))

_add_control('compliance slope',
             address=0x1C, length=1, nb_elem=2,
             models=('AX-12', 'AX-18', 'RX-24', 'RX-28', 'RX-64'))

_add_control('goal position',
             address=0x1E,
             dxl_to_si=conv.dxl_to_degree,
             si_to_dxl=conv.degree_to_dxl)

_add_control('moving speed',
             address=0x20,
             dxl_to_si=conv.dxl_to_speed,
             si_to_dxl=conv.speed_to_dxl)

_add_control('torque limit',
             address=0x22,
             dxl_to_si=conv.dxl_to_torque,
             si_to_dxl=conv.torque_to_dxl)

_add_control('goal position speed load',
             address=0x1E, nb_elem=3,
             dxl_to_si=lambda value, model: (conv.dxl_to_degree(value[0], model),
                                             conv.dxl_to_speed(value[1], model),
                                             conv.dxl_to_load(value[2], model)),
             si_to_dxl=lambda value, model: (conv.degree_to_dxl(value[0], model),
                                             conv.speed_to_dxl(value[1], model),
                                             conv.torque_to_dxl(value[2], model)))

_add_control('present position',
             address=0x24,
             access=_DxlAccess.readonly,
             dxl_to_si=conv.dxl_to_degree)

_add_control('present speed',
             address=0x26,
             access=_DxlAccess.readonly,
             dxl_to_si=conv.dxl_to_speed)

_add_control('present load',
             address=0x28,
             access=_DxlAccess.readonly,
             dxl_to_si=conv.dxl_to_load)

_add_control('present position speed load',
             address=0x24, nb_elem=3,
             access=_DxlAccess.readonly,
             dxl_to_si=lambda value, model: (conv.dxl_to_degree(value[0], model),
                                             conv.dxl_to_speed(value[1], model),
                                             conv.dxl_to_load(value[2], model)))

_add_control('present voltage',
             address=0x2A, length=1,
             access=_DxlAccess.readonly,
             dxl_to_si=conv.dxl_to_voltage)

_add_control('present temperature',
             address=0x2B, length=1,
             access=_DxlAccess.readonly,
             dxl_to_si=conv.dxl_to_temperature)

_add_control('moving',
             address=0x2E, length=1,
             access=_DxlAccess.readonly,
             dxl_to_si=conv.dxl_to_bool,
             getter_name='is_moving')

_add_control('punch',
             address=0x30, length=2,
             models=('AX-12', 'AX-18', 'RX-24', 'RX-28', 'RX-64'))



import sys
import time
import itertools
import threading

import PyQt4.QtCore
import PyQt4.QtGui
import PyQt4.uic

from collections import defaultdict
from pkg_resources import resource_filename

import pypot.dynamixel


# MARK: - DxlIO connection

__dxl_io = None
__lock = threading.Lock()


def get_dxl_connection(port, baudrate, protocol="MX"):
    global __dxl_io

    __lock.acquire()
    if protocol == "XL":
        __dxl_io = pypot.dynamixel.Dxl320IO(port, baudrate, use_sync_read=False)
    else:
        __dxl_io = pypot.dynamixel.DxlIO(port, baudrate, use_sync_read=False)
    return __dxl_io


def release_dxl_connection():
    __dxl_io.close()
    __lock.release()


class HerboristApp(PyQt4.QtGui.QApplication):
    def __init__(self, argv):
        PyQt4.QtGui.QApplication.__init__(self, argv)
        self.window = PyQt4.uic.loadUi(resource_filename('pypot', '/tools/herborist/herborist.ui'))

        self.enable_motor_view(False)

        # slider / lcd connections
        self.window.rdt_slider.valueChanged.connect(self.window.rdt_lcd.display)
        self.window.torque_max_slider.valueChanged.connect(self.window.torque_max_lcd.display)
        self.window.lower_limit_dial.valueChanged.connect(self.window.lower_limit_lcd.display)
        self.window.upper_limit_dial.valueChanged.connect(self.window.upper_limit_lcd.display)
        self.window.present_position_dial.valueChanged.connect(self.window.present_position_lcd.display)
        self.window.goal_position_dial.valueChanged.connect(self.window.goal_position_lcd.display)

        # Connect the button to their cb
        self.window.scan_button.pressed.connect(self.start_scanning)
        self.window.abort_button.pressed.connect(self.abort_scanning)

        # We manually uncheck most baudrates in the selection list
        # as unchecking them in the ui editor make the checkbox goes away
        for i in range(1, len(self.window.baudrate_list)):
            self.window.baudrate_list.item(i).setCheckState(False)

        # We prevent the minimum id to be greater than the maximum id
        def min_id_changed(self, new_id):
            if new_id >= self.window.max_id.value():
                self.window.min_id.setValue(self.window.max_id.value())

        def max_id_changed(self, new_id):
            if new_id <= self.window.min_id.value():
                self.window.max_id.setValue(self.window.min_id.value())

        self.window.min_id.valueChanged.connect(lambda i: min_id_changed(self, i))
        self.window.max_id.valueChanged.connect(lambda i: max_id_changed(self, i))

        # We automatically refresh the port list
        self.refresh_port_thread = self.UpdatePortThread()
        self.refresh_port_thread.port_updated.connect(self.update_port)
        self.refresh_port_thread.start()

        # Motor View
        self.window.motor_tree.itemSelectionChanged.connect(self.update_motor_view)
        self.window.goal_position_dial.valueChanged.connect(self.update_motor_position)
        self.window.update_eeprom_button.pressed.connect(self.update_eeprom)
        self.window.torque_checkbox.stateChanged.connect(self.switch_torque)

        self.refresh_motor_thread = None

    # MARK: - Port update

    class UpdatePortThread(PyQt4.QtCore.QThread):
        port_updated = PyQt4.QtCore.pyqtSignal(list)

        def __init__(self):
            PyQt4.QtCore.QThread.__init__(self)
            self.daemon = True

        def run(self):
            while True:
                new_ports = pypot.dynamixel.get_available_ports()
                self.port_updated.emit(new_ports)
                time.sleep(0.1)

    def update_port(self, new_ports):
        old_ports = []
        for i in range(len(self.window.port_box)):
            old_ports.append(str(self.window.port_box.itemText(i)))

        selected_port = self.window.port_box.currentText()

        if old_ports != new_ports:
            self.window.port_box.clear()
            self.window.port_box.addItems(new_ports)

            for i in range(len(self.window.port_box)):
                if self.window.port_box.itemText(i) == selected_port:
                    self.window.port_box.setCurrentIndex(i)
                    break

            if old_ports and not new_ports:
                self.window.scan_button.setEnabled(False)
            elif not old_ports and new_ports:
                self.window.scan_button.setEnabled(True)

    # MARK: - Motor Scan
    def update_motor_tree(self, baud_for_ids):
        self.window.motor_tree.clear()

        for b, ids in baud_for_ids.iteritems():
            baud_root = PyQt4.QtGui.QTreeWidgetItem(self.window.motor_tree, [str(b)])
            baud_root.setExpanded(True)
            f = int(baud_root.flags()) - int(PyQt4.QtCore.Qt.ItemIsSelectable)
            baud_root.setFlags(PyQt4.QtCore.Qt.ItemFlags(f))

            dxl_io = get_dxl_connection(self.port, b, self.protocol)
            models = dxl_io.get_model(ids)
            release_dxl_connection()

            for id, model in zip(ids, models):
                PyQt4.QtGui.QTreeWidgetItem(baud_root, ['', str(id), model])

    def start_scanning(self):
        self.window.scan_button.setEnabled(False)
        self.window.motor_tree.setEnabled(False)
        self.window.abort_button.setEnabled(True)

        self.window.motor_tree.clear()

        baudrates = []
        for i in range(len(self.window.baudrate_list)):
            checkbox = self.window.baudrate_list.item(i)
            if checkbox.checkState() == 2:
                b = int(checkbox.text())
                if b == 57600 and self.usb_device == "USB2DYNAMIXEL":
                    b = 57142
                baudrates.append(b)

        id_range = range(self.window.min_id.value(), self.window.max_id.value() + 1)

        self.window.scan_progress.setValue(0)
        self.window.scan_progress.setMaximum(len(baudrates) * len(id_range) - 1)

        self.scan_thread = self.ScanThread(self.port, baudrates, self.protocol, id_range,
                                           self.window.motor_tree, self.window.scan_progress)
        self.scan_thread.done.connect(self.done_scanning)
        self.scan_thread.part_done.connect(self.window.scan_progress.setValue)
        self.scan_thread.start()

    def abort_scanning(self):
        self.scan_thread.abort()

    def done_scanning(self):
        self.window.motor_tree.setEnabled(True)
        self.window.abort_button.setEnabled(False)
        self.window.scan_button.setEnabled(True)

    class ScanThread(PyQt4.QtCore.QThread):
        done = PyQt4.QtCore.pyqtSignal()
        part_done = PyQt4.QtCore.pyqtSignal(int)

        def __init__(self, port, baudrates, protocol, id_range,
                     motor_tree, scan_progress):

            PyQt4.QtCore.QThread.__init__(self)
            self.daemon = True

            self.running = threading.Event()
            self.running.set()

            self.port = port
            self.baudrates = baudrates
            self.protocol = protocol
            self.id_range = id_range

            self.motor_tree = motor_tree
            self.scan_progress = scan_progress

        def run(self):
            for b in self.baudrates:
                baud_root = PyQt4.QtGui.QTreeWidgetItem(self.motor_tree, [str(b)])
                baud_root.setExpanded(True)
                f = int(baud_root.flags()) - int(PyQt4.QtCore.Qt.ItemIsSelectable)
                baud_root.setFlags(PyQt4.QtCore.Qt.ItemFlags(f))

                dxl_io = get_dxl_connection(self.port, b, self.protocol)

                for id in self.id_range:
                    if not self.running.is_set():
                        break

                    if dxl_io.ping(id):
                        model = dxl_io.get_model((id, ))[0]
                        PyQt4.QtGui.QTreeWidgetItem(baud_root, ['', str(id), model])

                    self.part_done.emit(self.scan_progress.value() + 1)

                release_dxl_connection()

            self.done.emit()

        def abort(self):
            self.scan_progress.setValue(0)
            self.running.clear()

    # MARK: - Motor View
    def update_motor_view(self):
        if self.refresh_motor_thread:
            self.refresh_motor_thread.stop()
            self.refresh_motor_thread.wait()

        if not self.selected_motors:
            self.enable_motor_view(False)
            return

        self.enable_motor_view(True)

        if len(list(itertools.chain(*self.selected_motors.values()))) > 1:
            for widget_name in ('id_spinbox',
                                'present_position_dial',
                                'goal_position_dial',
                                'present_position_lcd',
                                'goal_position_lcd'):
                getattr(self.window, widget_name).setEnabled(False)

        dxl_io = get_dxl_connection(self.port, self.baudrate, self.protocol)
        srl = dxl_io.get_status_return_level((self.id, ), convert=False)[0]
        if srl != 0:
            model = dxl_io.get_model((self.id, ))[0]
            rdt = dxl_io.get_return_delay_time((self.id, ))[0]
            firmware = dxl_io.get_firmware((self.id, ))[0]
            torque_max = dxl_io.get_max_torque((self.id, ))[0]
            angle_limit = dxl_io.get_angle_limit((self.id, ))[0]
            goal_pos = dxl_io.get_goal_position((self.id, ))[0]
            torque_enable = dxl_io.is_torque_enabled((self.id, ))[0]
        release_dxl_connection()

        self.window.id_spinbox.setValue(self.id)

        for i in range(len(self.window.baud_combobox)):
            if int(self.window.baud_combobox.itemText(i)) == self.baudrate:
                self.window.baud_combobox.setCurrentIndex(i)
                break

        self.window.srl_combobox.setCurrentIndex(srl)

        if srl != 0:
            self.window.rdt_slider.setValue(rdt)
            self.window.firmware_label.setText(str(firmware))
            self.window.torque_max_slider.setValue(torque_max)

            limit = 180 if model.startswith('MX') else 150
            for widget_name in ('lower_limit_dial', 'upper_limit_dial',
                                'present_position_dial', 'goal_position_dial'):
                getattr(self.window, widget_name).setMinimum(-limit)
                getattr(self.window, widget_name).setMaximum(limit)

            self.window.lower_limit_dial.setValue(angle_limit[0])
            self.window.upper_limit_dial.setValue(angle_limit[1])
            self.window.goal_position_dial.setValue(goal_pos)
            self.window.torque_checkbox.setChecked(torque_enable)

        self.refresh_motor_thread = self.UpdateMotorThread(self.port, self.baudrate, self.protocol, self.id)
        self.refresh_motor_thread.position_updated.connect(self.motor_position_updated)
        self.refresh_motor_thread.start()

    def update_motor_position(self, pos):
        if self.window.torque_checkbox.checkState():
            dxl_io = get_dxl_connection(self.port, self.baudrate, self.protocol)
            dxl_io.set_goal_position({self.id: pos})
            time.sleep(0.05)
            release_dxl_connection()

    def motor_position_updated(self, pos):
        self.window.present_position_dial.setValue(pos)

    def switch_torque(self, torque_enable):
        if torque_enable:
            self.window.goal_position_dial.setEnabled(True)
            self.window.goal_position_lcd.setEnabled(True)
        else:
            self.window.goal_position_dial.setEnabled(False)
            self.window.goal_position_lcd.setEnabled(False)

        for b, ids in self.selected_motors.iteritems():
            dxl_io = get_dxl_connection(self.port, b, self.protocol)
            (dxl_io.enable_torque if torque_enable else dxl_io.disable_torque)(ids)
            time.sleep(0.05)
            release_dxl_connection()

    def enable_motor_view(self, enabled):
        for i in range(len(self.window.motor_layout)):
            self.window.motor_layout.itemAt(i).widget().setEnabled(enabled)

    def update_eeprom(self):
        if self.refresh_motor_thread:
            self.refresh_motor_thread.stop()
            self.refresh_motor_thread.wait()

        rdt = self.window.rdt_slider.value()
        srl = self.window.srl_combobox.currentIndex()
        torque_max = self.window.torque_max_slider.value()
        lower_limit = self.window.lower_limit_dial.value()
        upper_limit = self.window.upper_limit_dial.value()
        new_id = self.window.id_spinbox.value()

        for b, ids in self.selected_motors.iteritems():
            dxl_io = get_dxl_connection(self.port, b, self.protocol)
            dxl_io.set_return_delay_time(dict(zip(ids, itertools.repeat(rdt))))
            dxl_io.set_status_return_level(dict(zip(ids, itertools.repeat(srl))), convert=False)
            dxl_io.set_max_torque(dict(zip(ids, itertools.repeat(torque_max))))
            dxl_io.set_angle_limit(dict(zip(ids, itertools.repeat((lower_limit, upper_limit)))))
            time.sleep(0.15)
            release_dxl_connection()

        new_ids = self.ids.copy()

        old_ids = self.selected_motors.values()
        if len(old_ids) == 1 and len(old_ids[0]) == 1 and new_id != old_ids[0][0]:
            b, old_id = self.selected_motors.keys()[0], old_ids[0][0]
            dxl_io = get_dxl_connection(self.port, b, self.protocol)
            try:
                dxl_io.change_id({old_id: new_id})

                l = new_ids[b]
                l[l.index(old_id)] = new_id
                new_ids[b] = l

            except ValueError:
                PyQt4.QtGui.QMessageBox.about(self.window,
                                              '',
                                              'This id is already used.')
                pass

            release_dxl_connection()

        for b, ids in self.selected_motors.iteritems():
            new_b = int(self.window.baud_combobox.currentText())
            if b != new_b:
                dxl_io = get_dxl_connection(self.port, b, self.protocol)
                dxl_io.change_baudrate(dict(zip(ids, itertools.repeat(new_b))))
                time.sleep(0.1)
                release_dxl_connection()

            for id in ids:
                new_ids[new_b].append(id)
                new_ids[b].remove(id)
                new_ids[new_b].sort()

        self.update_motor_tree(new_ids)

    class UpdateMotorThread(PyQt4.QtCore.QThread):
        position_updated = PyQt4.QtCore.pyqtSignal(int)

        def __init__(self, port, baudrate, protocol, mid):
            PyQt4.QtCore.QThread.__init__(self)
            self.running = threading.Event()
            self.running.set()

            self.port = port
            self.baudrate = baudrate
            self.protocol = protocol
            self.mid = mid

        def stop(self):
            self.running.clear()

        def run(self):
            while self.running.is_set():
                dxl_io = get_dxl_connection(self.port, self.baudrate, self.protocol)
                pos = dxl_io.get_present_position((self.mid, ))[0]
                release_dxl_connection()

                self.position_updated.emit(pos)

                time.sleep(0.05)

    # MARK: - Shortcut properties
    @property
    def port(self):
        return str(self.window.port_box.currentText())

    @property
    def protocol(self):
        return str(self.window.protocol_box.currentText())

    @property
    def usb_device(self):
        return str(self.window.usb_device_box.currentText())

    @property
    def baudrate(self):
        return self.selected_motors.keys()[0]

    @property
    def id(self):
        return self.selected_motors[self.baudrate][0]

    @property
    def selected_motors(self):
        motors = defaultdict(list)

        selected_items = self.window.motor_tree.selectedItems()
        selected_items = [item for item in selected_items if item.parent()]

        for item in selected_items:
            motors[int(item.parent().text(0))].append(int(item.text(1)))

        return motors

    @property
    def ids(self):
        ids = defaultdict(list)

        root = self.window.motor_tree.invisibleRootItem()
        for i in range(root.childCount()):
            baud_node = root.child(i)

            if baud_node.text(0):
                b = int(baud_node.text(0))
                ids[b]
            else:
                ids[0].append(int(baud_node.text(1)))

            for j in range(baud_node.childCount()):
                id_node = baud_node.child(j)
                ids[b].append(int(id_node.text(1)))

        return ids


def main():
    app = HerboristApp(sys.argv)
    app.window.show()
    sys.exit(app.exec_())

    __lock.acquire()


if __name__ == '__main__':
    main()



from __future__ import division

import time
import json

from threading import Timer

from tornado.ioloop import IOLoop
from tornado.websocket import WebSocketHandler
from tornado.web import Application

from .server import AbstractServer


class RepeatedTimer(object):
    def __init__(self, interval, function, *args, **kwargs):
        self._timer = None
        self.interval = interval
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.is_running = False
        self.start()

    def _run(self):
        self.is_running = False
        self.start()
        self.function(*self.args, **self.kwargs)

    def start(self):
        if not self.is_running:
            self._timer = Timer(self.interval, self._run)
            self._timer.start()
            self.is_running = True

    def stop(self):
        self._timer.cancel()
        self.is_running = False


class WsSocketHandler(WebSocketHandler):
    time_step = 1 / 30

    def check_origin(self, origin):
        return True

    def open(self):
        if not self.quiet:
            print('WebSocket connection open.')

        self.rt = RepeatedTimer(self.time_step, self.publish_robot_state)

    def on_close(self):
        if not self.quiet:
            print('WebSocket connection closed: {0}'.format(self.close_reason))
        self.rt.stop()

    def on_message(self, message):
        if not self.quiet:
            print('{}: Received {}'.format(time.time(), message))

        self.handle_command(json.loads(message))

    def publish_robot_state(self):
        state = {
            m.name: {
                'present_position': m.present_position,
                'present_speed': m.present_speed,
                'present_load': m.present_load,
                'led': m.led,
                'present_temperature': m.present_temperature,
            }
            for m in self.robot.motors
        }
        self.write_message(json.dumps(state))

    def handle_command(self, command):
        for motor, values in command.items():
            m = getattr(self.robot, motor)

            for register, value in values.items():
                setattr(m, register, value)


class WsRobotServer(AbstractServer):
    def __init__(self, robot, host='0.0.0.0', port='9009', quiet=True):
        AbstractServer.__init__(self, robot, host, port)
        WsSocketHandler.robot = robot
        WsSocketHandler.quiet = quiet

    def run(self):
        loop = IOLoop()
        app = Application([
            (r'/', WsSocketHandler)
        ])
        app.listen(self.port)
        loop.start()



import smbus
import threading


class I2cController(object):
    used_bus = {}

    def __init__(self, pin_number):
        if pin_number not in I2cController.used_bus:
            I2cController.used_bus[pin_number] = smbus.SMBus(pin_number)

        self.bus = I2cController.used_bus[pin_number]
        self.lock = threading.Lock()

    def read_byte_data(self, addr, cmd):
        with self.lock:
            return self.bus.read_byte_data(addr, cmd)

    def write_byte_data(self, addr, cmd, val):
        with self.lock:
            return self.bus.write_byte_data(addr, cmd, val)

    def read_word_data(self, addr, cmd):
        with self.lock:
            return self.bus.read_word_data(addr, cmd)

    def write_word_data(self, addr, cmd, val):
        with self.lock:
            return self.bus.write_word_data(addr, cmd, val)



import numpy

from copy import deepcopy
from collections import defaultdict

from .primitive import Primitive, LoopPrimitive


class Sinus(LoopPrimitive):
    """ Apply a sinus on the motor specified as argument. Parameters (amp, offset and phase) should be specified in degree. """
    properties = LoopPrimitive.properties + ['frequency', 'amplitude', 'offset', 'phase']

    def __init__(self, robot, refresh_freq,
                 motor_list,
                 amp=1, freq=0.5, offset=0, phase=0):

        LoopPrimitive.__init__(self, robot, refresh_freq)

        self._freq = freq
        self._amp = amp
        self._offset = offset
        self._phase = phase

        self.motor_list = [self.get_mockup_motor(m) for m in motor_list]

    def update(self):
        """ Compute the sin(t) where t is the elapsed time since the primitive has been started. """
        pos = self._amp * numpy.sin(self._freq * 2.0 * numpy.pi * self.elapsed_time +
                                    self._phase * numpy.pi / 180.0) + self._offset

        for m in self.motor_list:
            m.goal_position = pos

    @property
    def frequency(self):
        return self._freq

    @frequency.setter
    def frequency(self, new_freq):
        self._freq = new_freq

    @property
    def amplitude(self):
        return self._amp

    @amplitude.setter
    def amplitude(self, new_amp):
        self._amp = new_amp

    @property
    def offset(self):
        return self._offset

    @offset.setter
    def offset(self, new_offset):
        self._offset = new_offset

    @property
    def phase(self):
        return self._phase

    @phase.setter
    def phase(self, new_phase):
        self._phase = new_phase


class Cosinus(Sinus):
    """ Apply a cosinus on the motor specified as argument. Parameters (amp, offset and phase) should be specified in degree. """

    def __init__(self, robot, refresh_freq,
                 motor_list,
                 amp=1, freq=0.5, offset=0, phase=0):

        Sinus.__init__(self, robot, refresh_freq,
                       motor_list,
                       amp, freq, offset, phase=(numpy.pi / 2 + phase))


class PositionWatcher(LoopPrimitive):
    def __init__(self, robot, refresh_freq, watched_motors):
        LoopPrimitive.__init__(self, robot, refresh_freq)

        self._pos = defaultdict(list)
        self.watched_motors = watched_motors
        self._duration = 0.

    @property
    def record_positions(self):
        return deepcopy(self._pos)

    def setup(self):
        self._pos.clear()

    def update(self):
        for m in self.watched_motors:
            self._pos[m.name].append(m.present_position)

        self._duration = self.elapsed_time

    def plot(self, ax):
        for m, pos in self._pos.items():
            t = numpy.linspace(0, self._duration, len(pos))
            ax.plot(t, pos)
        ax.set_ylabel('position (in deg)')
        ax.set_xlabel('time (in s)')
        ax.legend(self._pos.keys(), loc='best')


class SimplePosture(Primitive):
    def __init__(self, robot, duration):
        Primitive.__init__(self, robot)

        self.duration = duration

    def setup(self):
        self._speeds = {m: m.moving_speed for m in self.robot.motors}

        if hasattr(self, 'leds'):
            for m, c in self.leds.items():
                m.led = c

    def run(self):
        if not hasattr(self, 'target_position'):
            raise NotImplementedError('You have to define "target_position" first!')

        for m in self.robot.motors:
            m.compliant = False

        self.robot.goto_position(self.target_position, self.duration, wait=True)

    def teardown(self):
        for m, s in self._speeds.items():
            m.moving_speed = s

        if hasattr(self, 'leds'):
            for m, c in self.leds.items():
                m.led = 'off'



import threading

from . import pypot_time as time


class StoppableThread(object):
    """ Stoppable version of python Thread.

    This class provides the following mechanism on top of "classical" python Thread:
        * you can stop the thread (if you defined your run method accordingly).
        * you can restart a thread (stop it and re-run it)
        * you can pause/resume a thread

    .. warning:: It is up to the subclass to correctly respond to the stop, pause/resume signals (see :meth:`~pypot.utils.stoppablethread.StoppableThread.run` for details).

    """
    def __init__(self, setup=None, target=None, teardown=None):
        """
        :param func setup: specific setup function to use (otherwise self.setup)
        :param func target: specific target function to use (otherwise self.run)
        :param func teardown: specific teardown function to use (otherwise self.teardown)

        """
        self._started = threading.Event()
        self._running = threading.Event()
        self._resume = threading.Event()

        self._setup = self.setup if setup is None else setup
        self._target = self.run if target is None else target
        self._teardown = self.teardown if teardown is None else teardown
        self._crashed = False

    def start(self):
        """ Start the run method as a new thread.

        It will first stop the thread if it is already running.

        """
        if self.running:
            self.stop()

        self._thread = threading.Thread(target=self._wrapped_target)
        self._thread.daemon = True
        self._thread.start()

    def stop(self, wait=True):
        """ Stop the thread.

        More precisely, sends the stopping signal to the thread. It is then up to the run method to correctly responds.

        """
        if self.started:
            self._running.clear()
            self._resume.set()

            # We cannot wait for ourself
            if wait and (threading.current_thread() != self._thread):
                while self._thread.is_alive():
                    self._running.clear()
                    self._resume.set()
                    self._thread.join(timeout=1.0)

            self._started.clear()
            self._resume.clear()

    def join(self):
        """ Wait for the thread termination. """
        if not self.started:
            raise RuntimeError('cannot join thread before it is started')
        self._thread.join()

    @property
    def running(self):
        """ Whether the thread is running. """
        return self._running.is_set()

    @property
    def started(self):
        """ Whether the thread has been started. """
        return self._started.is_set()

    def wait_to_start(self, allow_failure=False):
        """ Wait for the thread to actually starts. """
        self._started.wait()

        if self._crashed and not allow_failure:
            self._thread.join()
            raise RuntimeError('Setup failed, see {} Traceback'
                               'for details.'.format(self._thread.name))

    def should_stop(self):
        """ Signals if the thread should be stopped or not. """
        return not self.running

    def wait_to_stop(self):
        """ Wait for the thread to terminate. """
        if not self.started:
            self.wait_to_start()
        self.join()

    def setup(self):
        """ Setup method call just before the run. """
        pass

    def run(self):
        """ Run method of the thread.

        .. note:: In order to be stoppable (resp. pausable), this method has to check the running property - as often as possible to improve responsivness - and terminate when :meth:`~pypot.utils.stoppablethread.StoppableThread.should_stop` (resp. :meth:`~pypot.utils.stoppablethread.StoppableThread.should_pause`) becomes True.
            For instance::

                while self.should_stop():
                    do_atom_work()
                    ...

        """
        pass

    def teardown(self):
        """ Teardown method call just after the run. """
        pass

    def _wrapped_target(self):
        try:
            self._setup()

            self._started.set()
            self._resume.set()

            self._running.set()
            self._target()
            self._running.clear()

            self._teardown()
        # In case something goes wrong within the thread
        # we try/catch the exceptions
        # clear all Condition locks (to avoid blocking main thread)
        # and re-raise exception for backtrace
        except:
            self._crashed = True
            self._started.set()
            self._running.clear()
            self._resume.clear()
            raise

    def should_pause(self):
        """ Signals if the thread should be paused or not. """
        return self.paused

    @property
    def paused(self):
        return not self._resume.is_set()

    def pause(self):
        """ Requests the thread to pause. """
        self._resume.clear()

    def resume(self):
        """ Requests the thread to resume. """
        self._resume.set()

    def wait_to_resume(self):
        """ Waits until the thread is resumed. """
        self._resume.wait()


def make_update_loop(thread, update_func):
    """ Makes a run loop which calls an update function at a predefined frequency. """
    while not thread.should_stop():
        if thread.should_pause():
            thread.wait_to_resume()

        start = time.time()
        if hasattr(thread, '_updated'):
            thread._updated.clear()
        update_func()
        if hasattr(thread, '_updated'):
            thread._updated.set()
        end = time.time()

        dt = thread.period - (end - start)

        if dt > 0:
            time.sleep(dt)


class StoppableLoopThread(StoppableThread):
    """ LoopThread calling an update method at a pre-defined frequency.

    .. note:: This class does not mean to be accurate. The given frequency will be approximately followed - depending for instance on CPU load - and only reached if the update method takes less time than the chosen loop period.

    """
    def __init__(self, frequency, update=None):
        """
        :params float frequency: called frequency of the :meth:`~pypot.stoppablethread.StoppableLoopThread.update` method

        """
        StoppableThread.__init__(self)

        self.period = 1.0 / frequency
        self._update = self.update if update is None else update
        self._updated = threading.Event()

    def run(self):
        """ Called the update method at the pre-defined frequency. """
        make_update_loop(self, self._update)

    def update(self):
        """ Update method called at the pre-defined frequency. """
        pass



import time
import logging

from .io.abstract_io import DxlError
from ..robot.controller import MotorsController

logger = logging.getLogger(__name__)


class DxlController(MotorsController):
    def __init__(self, io, motors, sync_freq, synchronous,
                 mode, regname, varname=None):
        MotorsController.__init__(self, io, motors, sync_freq)

        self.ids = [m.id for m in self.working_motors]
        self.synchronous = synchronous

        self.mode = mode
        self.regname = regname
        self.varname = regname if varname is None else varname

        for m in motors:
            if mode == 'get':
                m._read_synchronous[self.varname] = self.synchronous
            else:
                m._write_synchronous[self.varname] = self.synchronous

    @property
    def working_motors(self):
        return [m for m in self.motors if not m._broken]

    @property
    def synced_motors(self):
        motors = [m for m in self.working_motors if self.varname in m.registers]

        if self.synchronous:
            motors = ([m for m in motors if m._read_synced[self.varname].needed]
                      if self.mode == 'get' else
                      [m for m in motors if m._write_synced[self.varname].needed])

        return motors

    def setup(self):
        if self.mode == 'set':
            MAX_TRIALS = 25
            for _ in range(MAX_TRIALS):
                if self.get_register(self.working_motors, disable_sync_read=True):
                    break
                time.sleep(0.1)
            else:
                raise IOError('Cannot initialize syncloop for "{}". You need to desactivate sync_read if you use a usb2dynamixel device. '.format(
                              self.regname))

    def update(self):
        if not self.synced_motors:
            return

        return (self.get_register(self.synced_motors)
                if self.mode == 'get' else
                self.set_register(self.synced_motors))

    def get_register(self, motors, disable_sync_read=False):
        """ Gets the value from the specified register and sets it to the :class:`~pypot.dynamixel.motor.DxlMotor`. """
        if not motors:
            return False

        ids = [m.id for m in motors]
        getter = getattr(self.io, 'get_{}'.format(self.regname))

        values = (sum([list(getter([id])) for id in ids], [])
                  if disable_sync_read else
                  getter(ids))

        if not values:
            return False

        for m, val in zip(motors, values):
            m.__dict__[self.varname] = val

        for m in motors:
            m._read_synced[self.varname].done()

        return True

    def set_register(self, motors):
        """ Gets the value from :class:`~pypot.dynamixel.motor.DxlMotor` and sets it to the specified register. """
        if not motors:
            return
        ids = [m.id for m in motors]

        values = (m.__dict__[self.varname] for m in motors)
        getattr(self.io, 'set_{}'.format(self.regname))(dict(zip(ids, values)))

        for m in motors:
            m._write_synced[self.varname].done()


class AngleLimitRegisterController(DxlController):
    def __init__(self, io, motors, sync_freq, synchronous):
        DxlController.__init__(self, io, motors, sync_freq,
                               synchronous, 'get', 'angle_limit')

        self.varnames = ['lower_limit', 'upper_limit']
        for m in motors:
            for var in self.varnames:
                m._read_synchronous[var] = self.synchronous

    @property
    def synced_motors(self):
        motors = self.working_motors

        if self.synchronous:
            sync_motors = []

            for m in motors:
                for var in self.varnames:
                    if m._read_synced[var].needed:
                        sync_motors.append(m)

            motors = sync_motors

        return motors

    def get_register(self, motors):
        if not motors:
            return

        ids = [m.id for m in motors]
        values = self.io.get_angle_limit(ids)

        for m, val in zip(motors, values):
            m.__dict__['lower_limit'], m.__dict__['upper_limit'] = val

        for m in motors:
            for var in ['lower_limit', 'upper_limit']:
                m._read_synced[var].done()


class PosSpeedLoadDxlController(DxlController):
    def __init__(self, io, motors, sync_freq):
        DxlController.__init__(self, io, motors, sync_freq,
                               False, 'get', 'present_position')

    def setup(self):
        torques = self.io.is_torque_enabled(self.ids)
        for m, c in zip(self.working_motors, torques):
            m.compliant = not c
        self._old_torques = torques

        try:
            values = self.io.get_goal_position_speed_load(self.ids)
            positions, speeds, loads = zip(*values)
        except ValueError:
            raise DxlError("Couldn't initialize pos/speed/load sync loop!")

        for m, p, s, l in zip(self.working_motors, positions, speeds, loads):
            m.__dict__['goal_position'] = p
            m.__dict__['moving_speed'] = s
            m.__dict__['torque_limit'] = l

    def update(self):
        self.get_present_position_speed_load(self.working_motors)
        self.set_goal_position_speed_load(self.working_motors)

    def get_present_position_speed_load(self, motors):
        ids = [m.id for m in motors]
        values = self.io.get_present_position_speed_load(ids)

        if not values:
            logger.warning('Timeout when getting pos/speed/load from %s', ids)
            return

        positions, speeds, loads = zip(*values)

        for m, p, s, l in zip(motors, positions, speeds, loads):
            m.__dict__['present_position'] = p
            m.__dict__['present_speed'] = s
            m.__dict__['present_load'] = l

    def set_goal_position_speed_load(self, motors):
        change_torque = {}
        torques = [not m.compliant for m in motors]
        for m, t, old_t in zip(motors, torques, self._old_torques):
            if t != old_t:
                change_torque[m.id] = t
        self._old_torques = torques
        if change_torque:
            self.io._set_torque_enable(change_torque)

        rigid_motors = [m for m in motors if not m.compliant]
        ids = tuple(m.id for m in rigid_motors)

        if not ids:
            return

        values = ((m.__dict__['goal_position'],
                   m.__dict__['moving_speed'],
                   m.__dict__['torque_limit']) for m in rigid_motors)
        self.io.set_goal_position_speed_load(dict(zip(ids, values)))



from .rest import RESTRobot


class AbstractServer(object):
    def __init__(self, robot, host, port):
        self.restful_robot = RESTRobot(robot)
        self.host, self.port = host, port

    def run(self):
        raise NotImplementedError


try:
    import zerorpc

    class RemoteRobotServer(AbstractServer):
        def run(self):
            server = zerorpc.Server(self.restful_robot)
            server.bind('tcp://{}:{}'.format(self.host, self.port))
            server.run()

except ImportError:
    pass



import sys

from .abstractcreature import AbstractPoppyCreature

module = sys.modules[__name__]


installed_poppy_creatures = {}
# Feel free to make a pull request to add your own creature here
existing_creatures = ['poppy-humanoid', 'poppy-torso', 'poppy-ergo-jr', 'poppy-4dof-arm-mini', 'poppy-6dof-right-arm', 'poppy-dragster-mini', 'poppy-ergo','roboticia-quattro','roboticia-starter', 'doggy']


for creature in existing_creatures:
    package = creature.replace('-', '_')
    cls_name = ''.join(x.capitalize() or '_' for x in package.split('_'))

    try:
        cls = getattr(__import__(package), cls_name)
        installed_poppy_creatures[creature] = cls
        setattr(module, cls_name, cls)

    except (ImportError, AttributeError):
        pass



# -*- coding: utf-8 -*-


"""
    This module describes all the conversion method used to transform value from the representation used by the dynamixel motor to a more standard form (e.g. degrees, volt...).

    For compatibility issue all comparison method should be written in the following form (even if the model is not actually used):
        * def my_conversion_from_dxl_to_si(value, model): ...
        * def my_conversion_from_si_to_dxl(value, model): ...

    .. note:: If the control is readonly you only need to write the dxl_to_si conversion.

    """

import numpy
import itertools

from enum import Enum

# MARK: - Position

position_range = {
    'MX': (4096, 360.0),
    '*': (1024, 300.0)
}

torque_max = {  # in N.m
    'MX-106': 8.4,
    'MX-64': 6.,
    'MX-28': 2.5,
    'MX-12': 1.2,
    'AX-12': 1.2,
    'AX-18': 1.8,
    'RX-24': 2.6,
    'RX-28': 2.5,
    'RX-64': 4.,
    'XL-320': 0.39,
}

velocity = {  # in degree/s
    'MX-106': 270.,
    'MX-64': 378.,
    'MX-28': 330.,
    'MX-12': 2820.,
    'AX-12': 354.,
    'AX-18': 582.,
    'RX-24': 756.,
    'RX-28': 402.,
    'RX-64': 294.,
}


def dxl_to_degree(value, model):
    model = 'MX' if model.startswith('MX') else '*'
    max_pos, max_deg = position_range[model]

    return round(((max_deg * float(value)) / (max_pos - 1)) - (max_deg / 2), 2)


def degree_to_dxl(value, model):
    model = 'MX' if model.startswith('MX') else '*'
    max_pos, max_deg = position_range[model]

    pos = int(round((max_pos - 1) * ((max_deg / 2 + float(value)) / max_deg), 0))
    pos = min(max(pos, 0), max_pos - 1)

    return pos

# MARK: - Speed


def dxl_to_speed(value, model):
    cw, speed = divmod(value, 1024)
    direction = (-2 * cw + 1)

    speed_factor = 0.114 if model.startswith('MX') else 0.111

    return direction * (speed * speed_factor) * 6


def speed_to_dxl(value, model):
    direction = 1024 if value < 0 else 0
    speed_factor = 0.114 if model.startswith('MX') else 0.111

    max_value = 1023 * speed_factor * 6
    value = min(max(value, -max_value), max_value)

    return int(round(direction + abs(value) / (6 * speed_factor), 0))

# MARK: - Torque


def dxl_to_torque(value, model):
    return round(value / 10.23, 1)


def torque_to_dxl(value, model):
    return int(round(value * 10.23, 0))


def dxl_to_load(value, model):
    cw, load = divmod(value, 1024)
    direction = -2 * cw + 1

    return dxl_to_torque(load, model) * direction

# PID Gains


def dxl_to_pid(value, model):
    return (value[0] * 0.004,
            value[1] * 0.48828125,
            value[2] * 0.125)


def pid_to_dxl(value, model):
    def truncate(x):
        return int(max(0, min(x, 254)))
    return [truncate(x * y) for x, y in zip(value, (250, 2.048, 8.0))]

# MARK: - Model


dynamixelModels = {
    12: 'AX-12',    # 12 + (0<<8)
    18: 'AX-18',    # 18 + (0<<8)
    24: 'RX-24',    # 24 + (0<<8)
    28: 'RX-28',    # 28 + (0<<8)
    29: 'MX-28',    # 29 + (0<<8)
    64: 'RX-64',    # 64 + (0<<8)
    360: 'MX-12',   # 104 + (1<<8)
    310: 'MX-64',   # 54 + (1<<8)
    320: 'MX-106',  # 64 + (1<<8)
    350: 'XL-320',  # 94 + (1<<8)
}


def dxl_to_model(value, dummy=None):
    return dynamixelModels[value]
# MARK: - Drive Mode


def check_bit(value, offset):
    return bool(value & (1 << offset))


def dxl_to_drive_mode(value, model):
    return ('reverse' if check_bit(value, 0) else 'normal',
            'slave' if check_bit(value, 1) else 'master')


def drive_mode_to_dxl(value, model):
    return (int('slave' in value) << 1 | int('reverse' in value))

# MARK: - Baudrate


dynamixelBaudrates = {
    1: 1000000.0,
    3: 500000.0,
    4: 400000.0,
    7: 250000.0,
    9: 200000.0,
    16: 117647.1,
    34: 57600.0,
    103: 19230.8,
    207: 9615.4,
    250: 2250000.0,
    251: 2500000.0,
    252: 3000000.0,
}


def dxl_to_baudrate(value, model):
    return dynamixelBaudrates[value]


def baudrate_to_dxl(value, model):
    for k, v in dynamixelBaudrates.iteritems():
        if (abs(v - value) / float(value)) < 0.05:
            return k
    raise ValueError('incorrect baudrate {} (possible values {})'.format(value, dynamixelBaudrates.values()))

# MARK: - Return Delay Time


def dxl_to_rdt(value, model):
    return value * 2


def rdt_to_dxl(value, model):
    return int(value / 2)

# MARK: - Temperature


def dxl_to_temperature(value, model):
    return float(value)


def temperature_to_dxl(value, model):
    return int(value)

# MARK: - Voltage


def dxl_to_voltage(value, model):
    return value * 0.1


def voltage_to_dxl(value, model):
    return int(value * 10)

# MARK: - Status Return Level


status_level = ('never', 'read', 'always')


def dxl_to_status(value, model):
    return status_level[value]


def status_to_dxl(value, model):
    if value not in status_level:
        raise ValueError('status "{}" should be chosen among {}'.format(value, status_level))
    return status_level.index(value)

# MARK: - Error

# TODO: depend on protocol v1 vs v2


dynamixelErrors = ['None Error',
                   'Instruction Error',
                   'Overload Error',
                   'Checksum Error',
                   'Range Error',
                   'Overheating Error',
                   'Angle Limit Error',
                   'Input Voltage Error']


def dxl_to_alarm(value, model):
    return decode_error(value)


def decode_error(error_code):
    bits = numpy.unpackbits(numpy.asarray(error_code, dtype=numpy.uint8))
    return tuple(numpy.array(dynamixelErrors)[bits == 1])


def alarm_to_dxl(value, model):
    if not set(value).issubset(dynamixelErrors):
        raise ValueError('should only contains error among {}'.format(dynamixelErrors))

    indices = [len(dynamixelErrors) - 1 - dynamixelErrors.index(e) for e in value]
    return sum(2 ** i for i in indices)


XL320LEDColors = Enum('Colors', 'off red green yellow '
                      'blue pink cyan white')


def dxl_to_led_color(value, model):
    return XL320LEDColors(value + 1).name


def led_color_to_dxl(value, model):
    value = getattr(XL320LEDColors, value).value - 1
    value = int(value) & 0b111
    return value


control_modes = {
    1: 'wheel',
    2: 'joint',
}


def dxl_to_control_mode(value, _):
    return control_modes[value]


def control_mode_to_dxl(mode, _):
    return (next((v for v, m in control_modes.items()
                  if m == mode), None))

# MARK: - Various utility functions


def dxl_to_bool(value, model):
    return bool(value)


def bool_to_dxl(value, model):
    return int(value)


def dxl_decode(data):
    if len(data) not in (1, 2):
        raise ValueError('try to decode incorrect data {}'.format(data))

    if len(data) == 1:
        return data[0]

    if len(data) == 2:
        return data[0] + (data[1] << 8)


def dxl_decode_all(data, nb_elem):
    if nb_elem > 1:
        data = list(itertools.izip(*([iter(data)] * (len(data) // nb_elem))))
        return tuple(map(dxl_decode, data))
    else:
        return dxl_decode(data)


def dxl_code(value, length):
    if length not in (1, 2):
        raise ValueError('try to code value with an incorrect length {}'.format(length))

    if length == 1:
        return (value, )

    if length == 2:
        return (value % 256, value >> 8)


def dxl_code_all(value, length, nb_elem):
    if nb_elem > 1:
        return list(itertools.chain(*(dxl_code(v, length) for v in value)))
    else:
        return dxl_code(value, length)
